import logging

from odoo import fields, models
from odoo.exceptions import UserError, ValidationError

from odoo.addons.saas_worker.util import version_gte


_logger = logging.getLogger(__name__)


class PaymentProvider(models.Model):
    _inherit = 'payment.provider' if version_gte('16.0') else 'payment.acquirer'

    stripe_publishable_key = fields.Char(required_if_provider=False)
    stripe_secret_key = fields.Char(required_if_provider=False)

    stripe_account_id = fields.Char(readonly=True, groups='base.group_system', copy=False)
    stripe_account_details_submitted = fields.Boolean(
        readonly=True, groups='base.group_system', copy=False
    )

    _sql_constraints = [
        ('unique_account_id', 'UNIQUE(stripe_account_id)', 'The stripe_account_id should be unique.')
    ]

    #=== CONSTRAINT METHODS ===#

    def _stripe_has_connected_account(self):
        """ Override of `payment_stripe` to determine what is a "connected account".

        We rely on the `stripe_account_id` and both the `stripe_publishable_key` and
        `stripe_secret_key` because we allow the user to set their own API credentials to bypass the
        connected account in payment requests. In that case, the user must be able to set the
        provider in test mode.

        Note: self.ensure_one()
        Note: This method is only ever called from >=15.0 code as both the views and the methods in
              which it is called are respectively loaded and defined only if the server is in
              version >=15.0. Therefore, the method should not crash because of the missing
              `stripe_account_id` field.

        :return: Whether the provider is linked to a connected Stripe account
        :rtype: bool
        """
        self.ensure_one()

        neither_key_is_set = not self.stripe_secret_key and not self.stripe_publishable_key
        if self.stripe_account_id and neither_key_is_set:
            return True
        else:
            return super()._stripe_has_connected_account()

    def _stripe_onboarding_is_ongoing(self):
        """ Override of `payment_stripe` to check if there is an ongoing onboarding to Stripe.

        We rely on the `_stripe_has_connected_account` method to check if an onboarding has been
        started, and on the `stripe_account_details_submitted` field to check if the onboarding is
        finished.

        Note: self.ensure_one()

        :return: Whether the provider is linked to an ongoing onboarding to Stripe.
        :rtype: bool
        """
        self.ensure_one()

        if self._stripe_has_connected_account() and not self.stripe_account_details_submitted:
            return True
        else:
            return super()._stripe_onboarding_is_ongoing()

    # === ACTION METHODS === #

    def action_stripe_connect_account(self, menu_id=None):
        """ Override of payment_stripe to update the onboarding status on-the-fly.

        Note: self.ensure_one()

        :param int menu_id: The menu from which the user started the onboarding step, as an
                            `ir.ui.menu` id
        """
        if not self._stripe_has_connected_account() or self._stripe_onboarding_is_ongoing():
            self._update_stripe_onboarding_status()
        else:  # The account is already onboarded but the provider is disabled.
            self.state = 'enabled'  # Simply enable the provider.
        return super().action_stripe_connect_account(menu_id)

    def _update_stripe_onboarding_status(self, connected_account=None):
        """ Check if the account details have been submitted and activate the provider if so.

        Note: self.ensure_one()

        :param dict connected_account: The Stripe connected account
        :return: None
        """
        self.ensure_one()

        if not self.stripe_account_id:  # The account is not yet onboarded or has been reset
            return

        try:
            if not connected_account:
                connected_account = self._stripe_fetch_or_create_connected_account()
        except ValidationError:
            _logger.exception("could not fetch the connected account")
        else:
            if connected_account.get('details_submitted'):  # The onboarding is finished
                write_values = {
                    'stripe_account_details_submitted': True,
                    # We consider that the account can receive online payments as soon as the
                    # 'details_submitted' attribute is 'true'. Ideally, we should base our heuristic
                    # on 'charges_enabled' but we found out that the first attribute update was
                    # notified much faster through the webhook than the second one (few seconds vs
                    # few minutes). Worst, to enable live charges, the user must first upload
                    # identity documents on the Stripe onboarding form. In the (hopefully) rare case
                    # when completing the onboarding does not lead to the enabling of live charges,
                    # the user should be told by Stripe that some actions are required on their
                    # account.
                    'state': 'enabled',
                }
                if version_gte('16.0'):
                    write_values['is_published'] = True
                self.write(write_values)

                if version_gte('16.0'):
                    try:
                        self.action_stripe_verify_apple_pay_domain()
                    except UserError:  # The account is onboarded in test environment.
                        _logger.info(
                            "Skipped verification of the Apple Pay domain because the account is "
                            "onboarded in test environment."
                        )  # No big deal, only Apple Pay will be unusable.

                if version_gte('saas~16.4'):
                    self.env['onboarding.onboarding.step'].sudo().action_validate_step_payment_provider()
                else:
                    self.sudo().company_id._mark_payment_onboarding_step_as_done()

    def action_stripe_reset_connected_account(self):
        """ Reset Stripe connected account.

        Reset the Stripe Connect account id and its charges enabled flag. Disable the provider if at
        least one of the secret or publishable keys is not set.

        Note: self.ensure_one()
        """
        self.ensure_one()

        state = 'disabled' if self._stripe_has_connected_account() else self.state
        write_values = {
            'stripe_account_id': False,
            'stripe_account_details_submitted': False,
            'state': state,
        }
        if version_gte('16.0'):
            write_values['is_published'] = False
        return self.write(write_values)

    # === BUSINESS METHODS - PAYMENT FLOW === #

    def _get_stripe_extra_request_headers(self):
        """ Override of payment_stripe to add the connected account id to the extra request headers.

        If the provider's secret key is set, no header is added to use the provider's account.

        :return: The extra request headers.
        :rtype: dict
        """
        extra_headers = super()._get_stripe_extra_request_headers()
        if self._stripe_has_connected_account():
            extra_headers['Stripe-Account'] = self.stripe_account_id
        return extra_headers

    # === BUSINESS METHODS - STRIPE CONNECT ONBOARDING === #

    def _stripe_fetch_or_create_connected_account(self):
        """ Override of payment_stripe to fetch the account from Stripe if the account id is set.

        :return: The connected account
        :rtype: dict
        """
        if not self.stripe_account_id:  # Non-existing or reset account.
            connected_account = super()._stripe_fetch_or_create_connected_account()
            self.stripe_account_id = connected_account['id']
        else:  # Existing account.
            connected_account = self._stripe_make_request(
                f'accounts/{self.stripe_account_id}', method='GET'
            )
        return connected_account

    def _stripe_prepare_connect_account_payload(self):
        """ Override of `payment_stripe` to add the required metadata to the payload.

        :return: The Stripe-formatted payload for the creation request
        :rtype: dict
        """
        return {
            **super()._stripe_prepare_connect_account_payload(),
            # Used by the proxy to forward account update notifications to the database.
            'metadata[webhook_url]': self._get_stripe_webhook_url(),
        }

    def _stripe_prepare_proxy_data(self, **kwargs):
        """ Override of `payment_stripe` to add the required proxy data.

        :return: The Stripe-formatted payload for the creation request.
        :rtype: dict
        """
        return {
            **super()._stripe_prepare_proxy_data(**kwargs),
            # Used by the proxy to select the API key based on the platform account.
            'platform_account': self._stripe_select_platform_account(),
        }

    def _stripe_select_platform_account(self):
        """ Select and return the appropriate platform account for the provider.

        The platform account is selected based on the main currency of the provider's company's
        country. For providers charging mostly in USD, the account is onboarded on our US platform
        account to avoid incurring a conversion fee on payouts, otherwise the account is onboarded
        on our BE platform account.

        Note: If the main currency of the provider's company's country is changed, the account is
              reset to prevent making requests with the wrong platform account's API keys.
        Note: self.ensure_one()

        :return: The selected platform account, 'be' or 'us'.
        :rtype: str
        """
        return 'be' if self.company_id.currency_id.name != 'USD' else 'us'

    # === HELPER METHODS === #

    def _get_code(self):
        """ Return the `code` or `provider` field depending on whether the database is running in
        Odoo 16.0 or lower.

        :return: The provider code.
        :rtype: str
        """
        if version_gte('16.0'):
            return self.code
        else:
            return self.provider
